#!/usr/bin/env python3

from tests.utils import MLRTestCase


class TestMLR(MLRTestCase):
    """Test all multi-line-replacer (mlr) functionality"""

    def test_literal_replacement(self) -> None:
        """Should perform a literal textual replacement"""
        self.assert_file_replace(
            input_filenames=["input/test.editorconfig"],
            rule_filenames=["rules/editorconfig.md"],
            output_filenames=["output/test.editorconfig"],
            expected_cli_message="1 file changed, 0 files unchanged",
        )

    def test_literal_target_misindented(self) -> None:
        """
        Should perform a literal textual replacement even if target text is
        misindented
        """
        self.assert_file_replace(
            input_filenames=["input/test.editorconfig"],
            rule_filenames=["rules/editorconfig-misindented.md"],
            output_filenames=["output/test.editorconfig"],
            expected_cli_message="1 file changed, 0 files unchanged",
        )

    def test_match_until_end_of_line(self) -> None:
        """Should perform a replacement with MATCH_UNTIL_END_OF_LINE"""
        self.assert_file_replace(
            input_filenames=["input/lint.yml"],
            rule_filenames=["rules/ruff.md"],
            output_filenames=["output/lint-ruff.yml"],
            expected_cli_message="1 file changed, 0 files unchanged",
        )

    def test_match_all_between(self) -> None:
        """Should perform a replacement with MATCH_ALL_BETWEEN"""
        self.assert_file_replace(
            input_filenames=["input/lint.yml"],
            rule_filenames=["rules/python-version.md"],
            output_filenames=["output/lint-python-version.yml"],
            expected_cli_message="1 file changed, 0 files unchanged",
        )

    def test_normalize_indent_unit_in_replacement(self) -> None:
        """
        Should normalize indent unit in replacement to match indent unit of
        input text
        """
        self.assert_file_replace(
            input_filenames=["input/lint.yml"],
            rule_filenames=["rules/ruff-tab-indent.md"],
            output_filenames=["output/lint-ruff.yml"],
            expected_cli_message="1 file changed, 0 files unchanged",
        )

    def test_no_match(self) -> None:
        """
        Should leave the file untouched if no matches are found
        """
        self.assert_file_replace(
            input_filenames=["input/publish.yml"],
            rule_filenames=["rules/ruff.md"],
            output_filenames=["input/publish.yml"],
            expected_cli_message="1 file unchanged (no replacements made)",
        )

    def test_missing_code_blocks(self) -> None:
        """
        Should raise a RuntimeError if there are an odd number of code blocks
        """
        with self.assertRaises(RuntimeError):
            self.assert_file_replace(
                input_filenames=["input/publish.yml"],
                rule_filenames=["rules/missing-code-blocks.md"],
                output_filenames=["input/publish.yml"],
            )



#!/usr/bin/env python3.13

import argparse

from mlr.core import ExpandedPath, extract_code_blocks, replace_text


def pluralize(singular: str, plural: str, count: int) -> str:
    """
    Return "1 <singular>" or "<count> <plural>", where the noun is either in
    singular or plural form depending on the supplied count
    """
    if count == 1:
        return f"{count} {singular}"
    else:
        return f"{count} {plural}"


def get_cli_args() -> argparse.Namespace:
    """Define and parse CLI arguments"""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "input_paths",
        metavar="INPUT_FILE",
        nargs="+",
        type=ExpandedPath,
        help="One or more paths to files to apply replacements to.",
    )
    parser.add_argument(
        "-r",
        "--rules",
        metavar="RULE_FILE",
        dest="rule_paths",
        nargs="+",
        required=True,
        type=ExpandedPath,
        help="One or more paths to replacement rule Markdown files. Each file should contain pairs of triple-backtick (```) fenced code blocks, where the first fenced block is the text to be replaced and the second fenced block is the replacement text.",  # noqa: E501
    )
    return parser.parse_args()


def print_replacement_summary(total_file_count: int, total_files_changed: int) -> None:
    """
    Print a summary of how many files have been changed and how many
    replacements have been made
    """
    if total_files_changed:
        print(
            f"{pluralize('file', 'files', total_files_changed)} changed, {pluralize('file', 'files', total_file_count - total_files_changed)} unchanged"  # noqa: E501
        )
    else:
        print(
            f"{pluralize('file', 'files', total_file_count - total_files_changed)} unchanged (no replacements made)"  # noqa: E501
        )


def main() -> None:
    """The entry point for the `multi-line-replacer` / `mlr` CLI program"""
    args = get_cli_args()
    total_files_changed = 0
    for input_path in args.input_paths:
        orig_input_text = input_path.read_text()
        input_text = orig_input_text
        # Apply each replacement rule to each input file
        for rule_path in args.rule_paths:
            rule_text = rule_path.read_text()
            code_blocks = extract_code_blocks(rule_text)
            # Enumerate fenced code blocks in pairs to get each pair of
            # target/replacement rules
            for target_text, replacement_text in zip(
                code_blocks[0::2], code_blocks[1::2]
            ):
                input_text = replace_text(input_text, target_text, replacement_text)
        if orig_input_text != input_text:
            total_files_changed += 1
        input_path.write_text(input_text)
    print_replacement_summary(
        total_file_count=len(args.input_paths),
        total_files_changed=total_files_changed,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3.13
import os
import re
import sys
from pathlib import PosixPath, WindowsPath
from typing import Optional

# Detect proper Path subclass to inherit from based on the user's platform,
# since the top-level Path subclass is not designed to be subclassed directly
if sys.platform == "win32":
    BasePath = WindowsPath
else:
    BasePath = PosixPath


class ExpandedPath(BasePath):
    """
    Path subclass that automatically expands the user's home directory (i.e. ~)
    """

    def __new__(cls, path: str, **kwargs: object) -> "ExpandedPath":
        return super().__new__(cls, os.path.expanduser(path), **kwargs)


def extract_code_blocks(md_text: str) -> list[str]:
    """
    Extract contents of all fenced code blocks from the given Markdown text into
    a list, where each list item represents the contents of that code block
    (minus any optional language specifier)
    """
    code_blocks = re.findall(
        r"```(?:[\w\-]*)\n(.*?)\n```",
        md_text,
        flags=re.DOTALL,
    )
    # If number of code blocks is not even
    if len(code_blocks) % 2 != 0:
        raise RuntimeError(
            "Replacement file must have an even number of fenced code blocks."
        )
    return code_blocks


def get_indent_unit(text: str) -> Optional[str]:
    """
    Find the string representing the base unit of indentation in a given string
    of text; this can be either two spaces, four spaces, or a tab character
    """
    for indent_level in ((" " * 2), (" " * 4), "\t"):
        # The positive lookahead is because since 2-space indent is really a
        # subset of 4-space indent, we need to be able to distinguish between
        # the two, so we check to see if there
        if re.search(rf"^{indent_level}(?=\S)", text, flags=re.MULTILINE):
            return indent_level
    return None


def evaluate_wildcard_variables(text: str) -> str:
    """
    Evaluate textual wildcard variables in the given target text to achieve
    certain behaviors (like wildcard-matching through the end of a line, or a
    wildcard-match for only a single word)
    """
    wildcard_evaluations = {
        # Match all non-newline characters until the end of the line is reached
        "MATCH_UNTIL_END_OF_LINE": r"[^\n]*",
        # Match all non-newline characters between two delimiters (like quotes)
        "MATCH_ALL_BETWEEN": r"[^\n]*?",
    }
    for wildcard_var_name, replacement in wildcard_evaluations.items():
        text = re.sub(
            re.escape(wildcard_var_name),
            replacement,
            text,
        )
    return text


def replace_text(input_text: str, target_text: str, replacement_text: str) -> str:
    """
    Replace the given text in the input text with the replacement text,
    preserving indentation; the return value is the full output text after
    replacements have been made
    """
    replace_this_patt = "\n".join(
        (
            # Evaluate special wildcard variables like
            # MATCH_UNTIL_END_OF_LINE
            evaluate_wildcard_variables(rf"([ \t]*){re.escape(line.strip())}")
            if line
            else ""
        )
        for line in target_text.splitlines()
    )
    # Retrieve the base indentation level in the target text to ensure that the
    # replacement text is indented the same amount
    base_indent_matches = re.search(replace_this_patt, input_text)
    if not base_indent_matches:
        return input_text
    base_indent_level = base_indent_matches.group(1)
    # Ensure that the replacement text's preferred indentation unit matches that
    # of the input text
    input_indent_unit = get_indent_unit(input_text)
    replacement_indent_unit = get_indent_unit(replacement_text)
    if replacement_indent_unit and input_indent_unit:
        replacement_text = re.sub(
            replacement_indent_unit,
            input_indent_unit,
            replacement_text,
        )
    # Ensure that the replacement text is indented to the same amount as the
    # target text it is replacing
    replacement_text = "\n".join(
        # The ternary syntax is to prevent trailing whitespace from being added
        # to blank lines
        base_indent_level + line if line else ""
        for line in replacement_text.splitlines()
    )
    return re.sub(replace_this_patt, replacement_text, input_text)
